# -*- coding: utf-8 -*-

"""BoardGameGeek rankings spider."""

import os
import re

from datetime import datetime, timezone
from random import randint

from pytility import normalize_space, parse_date, parse_int
from scrapy import Request, Spider

from ..items import GameItem
from ..loaders import GameLoader
from ..utils import extract_bgg_id, now, parse_url

DIGITS_REGEX = re.compile(r"^\D*(\d+).*$")
BGG_URL_REGEX = re.compile(r"^.*(https?://?(www\.)?boardgamegeek\.com.*)$")
HTTP_REGEX = re.compile(r"^(https?):/([^/])")
DATE_PATH_REGEX = re.compile(r"^/[^/]+/(\d+).*$")
WEB_ARCHIVE_DATE_FORMAT = "%Y%m%d%H%M%S"


def _parse_int(element, xpath=None, css=None, default=None, lenient=False):
    if not element or (not xpath and not css):
        return default

    selected = element.xpath(xpath) if xpath else element.css(css)
    string = normalize_space(selected.extract_first())

    if not string:
        return default

    result = parse_int(string)

    if result is None and lenient:
        match = DIGITS_REGEX.match(string)
        result = parse_int(match.group(1)) if match else None

    return result if result is not None else default


def _extract_bgg_id(url):
    url = parse_url(
        url,
        (
            "boardgamegeek.com",
            "www.boardgamegeek.com",
            "archive.org",
            "web.archive.org",
        ),
    )

    bgg_id = extract_bgg_id(url)
    if bgg_id:
        return bgg_id

    match = BGG_URL_REGEX.match(url.path)

    if not match:
        return None

    url = HTTP_REGEX.sub(r"\1://\2", match.group(1))
    return extract_bgg_id(url)


def _parse_date(date, tzinfo=timezone.utc, format_str=WEB_ARCHIVE_DATE_FORMAT):
    try:
        date = datetime.strptime(date, format_str)
        return date.replace(tzinfo=tzinfo)
    except Exception:
        pass

    try:
        return parse_date(date, tzinfo, format_str)
    except Exception:
        pass

    return None


def _extract_date(url, tzinfo=timezone.utc, format_str=WEB_ARCHIVE_DATE_FORMAT):
    url = parse_url(url, ("archive.org", "web.archive.org",),)

    if not url:
        return None

    match = DATE_PATH_REGEX.match(url.path)

    return _parse_date(match.group(1), tzinfo, format_str) if match else None


class BggRankingsSpider(Spider):
    """BoardGameGeek rankings spider."""

    name = "bgg_rankings"
    allowed_domains = ("boardgamegeek.com", "archive.org")
    bgg_paths = (
        "browser.php?itemtype=game&sortby=rank",
        "rankbrowse.php3",
        "browse/boardgame",
        "top50.htm",
        "top50.php3",
        "topn.php3?count=50",
    )
    bgg_urls = (
        tuple(f"http://boardgamegeek.com/{path}" for path in bgg_paths)
        + tuple(f"https://boardgamegeek.com/{path}" for path in bgg_paths)
        + tuple(f"http://www.boardgamegeek.com/{path}" for path in bgg_paths)
        + tuple(f"https://www.boardgamegeek.com/{path}" for path in bgg_paths)
    )
    start_urls = (
        tuple(f"https://web.archive.org/web/{{date}}/{url}" for url in bgg_urls)
        + bgg_urls
    )
    item_classes = (GameItem,)

    custom_settings = {
        "DOWNLOAD_DELAY": 0.5,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 8,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 4,
        "DELAYED_RETRY_ENABLED": True,
        "DELAYED_RETRY_HTTP_CODES": (202,),
        "DELAYED_RETRY_DELAY": 5.0,
        "AUTOTHROTTLE_HTTP_CODES": (429, 503, 504),
        "START_DATE": parse_date(os.getenv("START_DATE"), tzinfo=timezone.utc),
        "EARLIEST_DATE": parse_date(os.getenv("EARLIEST_DATE"), tzinfo=timezone.utc)
        or datetime(2000, 1, 1, tzinfo=timezone.utc),
        "LATEST_DATE": parse_date(os.getenv("LATEST_DATE"), tzinfo=timezone.utc),
    }

    def start_requests(self):
        """Generate start requests."""

        start_date = parse_date(self.settings.get("START_DATE"), tzinfo=timezone.utc)

        if not start_date:
            earliest_date = (
                parse_date(self.settings.get("EARLIEST_DATE"), tzinfo=timezone.utc)
                or now()
            )
            latest_date = (
                parse_date(self.settings.get("LATEST_DATE"), tzinfo=timezone.utc)
                or now()
            )
            start_date_ts = randint(earliest_date.timestamp(), latest_date.timestamp())
            start_date = datetime.fromtimestamp(start_date_ts, tz=timezone.utc)

        self.logger.info("Start date: %s", start_date)

        start_date_str = start_date.strftime(WEB_ARCHIVE_DATE_FORMAT)

        for start_url in self.start_urls:
            yield Request(
                url=start_url.format(date=start_date_str),
                callback=self.parse,
                priority=1,
            )

    def parse(self, response):
        """
        @url https://boardgamegeek.com/browse/boardgame
        @returns items 100 100
        @returns requests 12 12
        """

        scraped_at = now()
        published_at = (
            _extract_date(response.url)
            or response.meta.get("published_at")
            or scraped_at
        )

        for next_page in response.xpath(
            "//a[contains(@title, 'page')]/@href"
        ).extract():
            yield response.follow(
                url=next_page,
                callback=self.parse,
                priority=1,
                meta={"published_at": published_at, "max_retry_times": 10},
            )

        for row in response.css("table#collectionitems tr"):
            link = row.css("td.collection_objectname a::attr(href)").extract_first()
            link = response.urljoin(link)
            bgg_id = _extract_bgg_id(link)

            if not bgg_id:
                continue

            year = _parse_int(
                element=row,
                css="td.collection_objectname span.smallerfont.dull",
                lenient=True,
            )
            image_url = row.css(
                "td.collection_thumbnail img::attr(src)"
            ).extract_first()
            image_url = [response.urljoin(image_url)] if image_url else None

            ldr = GameLoader(
                item=GameItem(
                    bgg_id=bgg_id,
                    year=year,
                    image_url=image_url,
                    published_at=published_at,
                    scraped_at=scraped_at,
                ),
                selector=row,
                response=response,
            )

            ldr.add_css("rank", "td.collection_rank")
            ldr.add_css("name", "td.collection_objectname a")

            values = row.css("td.collection_bggrating").extract()
            if len(values) == 3:
                ldr.add_value("bayes_rating", values[0])
                ldr.add_value("avg_rating", values[1])
                ldr.add_value("num_votes", values[2])

            yield ldr.load_item()

        for row in response.css("div.simplebox table tr"):
            cells = row.xpath("td")

            if len(cells) != 3:
                continue

            link = cells[1].xpath("a/@href").extract_first()
            link = response.urljoin(link)
            bgg_id = _extract_bgg_id(link)

            if not bgg_id:
                continue

            ldr = GameLoader(
                item=GameItem(
                    bgg_id=bgg_id, published_at=published_at, scraped_at=scraped_at,
                ),
                selector=row,
                response=response,
            )

            ldr.add_xpath("rank", "td[1]")
            ldr.add_xpath("name", "td[2]")
            ldr.add_xpath("bayes_rating", "td[3]")

            yield ldr.load_item()

        for row in response.css("table.gamebrowser_table tr"):
            cells = row.xpath("td")

            if len(cells) < 5:
                continue

            link = cells[2].xpath("a/@href").extract_first()
            link = response.urljoin(link)
            bgg_id = _extract_bgg_id(link)

            if not bgg_id:
                continue

            image_url = row.xpath("td[2]//img/@src").extract_first()
            image_url = [response.urljoin(image_url)] if image_url else None

            ldr = GameLoader(
                item=GameItem(
                    bgg_id=bgg_id,
                    image_url=image_url,
                    published_at=published_at,
                    scraped_at=scraped_at,
                ),
                selector=row,
                response=response,
            )

            ldr.add_xpath("rank", "td[1]")
            ldr.add_xpath("name", "td[3]")
            ldr.add_xpath("bayes_rating", "td[4]")
            ldr.add_xpath("num_votes", "td[5]")

            yield ldr.load_item()

        # Parse Top 50 page: top50.htm, top50.php3, topn.php3?count=50
        for row in response.xpath(
            "//table[tr/td[h3 and contains(., 'Bayesian Average')]]/tr"
        ):
            cells = row.xpath("td")

            if len(cells) < 4:
                continue

            link = cells[1].xpath("a/@href").extract_first()
            link = response.urljoin(link)
            bgg_id = _extract_bgg_id(link)
            rank = _parse_int(cells[0], xpath="text()", lenient=True)

            if not bgg_id or not rank:
                continue

            ldr = GameLoader(
                item=GameItem(
                    bgg_id=bgg_id,
                    rank=rank,
                    published_at=published_at,
                    scraped_at=scraped_at,
                ),
                selector=row,
                response=response,
            )

            ldr.add_xpath("name", "td[2]")
            ldr.add_xpath("bayes_rating", "td[3]")
            ldr.add_xpath("num_votes", "td[4]")

            yield ldr.load_item()

        for anchor in response.xpath(
            "//div[@id = 'wm-ipp']//table//a[@title and @href]"
        ):
            if parse_date(anchor.xpath("@title").extract_first()):
                yield response.follow(
                    url=anchor.xpath("@href").extract_first(),
                    callback=self.parse,
                    priority=-1,
                    meta={"max_retry_times": 10},
                )



# -*- coding: utf-8 -*-

""" BoardGameGeek spider """

import re
import statistics

from functools import partial
from itertools import repeat
from urllib.parse import urlencode

from pytility import batchify, clear_list, normalize_space, parse_int
from scrapy import signals
from scrapy import Request, Spider
from scrapy.utils.misc import arg_to_iter
from scrapy.utils.project import get_project_settings

from ..items import GameItem, RatingItem, UserItem
from ..loaders import GameLoader, RatingLoader, UserLoader
from ..utils import (
    extract_bgg_id,
    extract_bgg_user_name,
    extract_item,
    extract_query_param,
    now,
)

DIGITS_REGEX = re.compile(r"^\D*(\d+).*$")


def _parse_int(element, xpath, default=None, lenient=False):
    if not element or not xpath:
        return default

    string = normalize_space(element.xpath(xpath).extract_first())

    if not string:
        return default

    result = parse_int(string)

    if result is None and lenient:
        match = DIGITS_REGEX.match(string)
        result = parse_int(match.group(1)) if match else None

    return result if result is not None else default


def _parse_player_count(poll):
    for result in poll.xpath("results"):
        numplayers = normalize_space(result.xpath("@numplayers").extract_first())
        players = parse_int(numplayers)

        if not players and numplayers.endswith("+"):
            players = parse_int(numplayers[:-1]) or -1
            players += 1

        if not players:
            continue

        votes_best = _parse_int(result, 'result[@value = "Best"]/@numvotes', 0)
        votes_rec = _parse_int(result, 'result[@value = "Recommended"]/@numvotes', 0)
        votes_not = _parse_int(
            result, 'result[@value = "Not Recommended"]/@numvotes', 0
        )

        yield players, votes_best, votes_rec, votes_not


def _parse_votes(poll, attr="value", enum=False):
    if not poll:
        return

    for i, result in enumerate(poll.xpath("results/result"), start=1):
        value = i if enum else _parse_int(result, "@" + attr, lenient=True)
        numvotes = _parse_int(result, "@numvotes", 0)

        if value is not None:
            yield from repeat(value, numvotes)


def _value_id(items, sep=":"):
    for item in arg_to_iter(items):
        value = item.xpath("@value").extract_first() or ""
        id_ = item.xpath("@id").extract_first() or ""
        yield f"{value}{sep}{id_}" if id_ else value


def _value_id_rank(items, sep=":"):
    for item in arg_to_iter(items):
        value = item.xpath("@friendlyname").extract_first() or ""
        value = value[:-5] if value and value.lower().endswith(" rank") else value
        id_ = item.xpath("@id").extract_first() or ""
        yield f"{value}{sep}{id_}" if id_ else value


class BggSpider(Spider):
    """ BoardGameGeek spider """

    name = "bgg"
    allowed_domains = ("boardgamegeek.com",)
    start_urls = (
        "https://boardgamegeek.com/browse/boardgame/",
        "https://boardgamegeek.com/browse/user/numreviews",
        "https://boardgamegeek.com/browse/user/numsessions",
    )
    item_classes = (GameItem, UserItem, RatingItem)
    state = None

    # https://www.boardgamegeek.com/wiki/page/BGG_XML_API2
    xml_api_url = "https://www.boardgamegeek.com/xmlapi2"
    page_size = 100

    custom_settings = {
        "DOWNLOAD_DELAY": 0.5,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 8,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 4,
        "DELAYED_RETRY_ENABLED": True,
        "DELAYED_RETRY_HTTP_CODES": (202,),
        "DELAYED_RETRY_DELAY": 5.0,
        "AUTOTHROTTLE_HTTP_CODES": (429, 503, 504),
        "PULL_QUEUE_ENABLED": True,
    }

    scrape_ratings = False
    scrape_collections = False
    scrape_users = False
    min_votes = 20

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        """ initialise spider from crawler """

        kwargs.pop("settings", None)
        spider = cls(*args, settings=crawler.settings, **kwargs)
        spider._set_crawler(crawler)

        crawler.signals.connect(spider._spider_opened, signal=signals.spider_opened)

        return spider

    def __init__(self, *args, settings=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._ids_seen = set()

        settings = settings or get_project_settings()

        self.scrape_ratings = settings.getbool("SCRAPE_BGG_RATINGS")
        self.scrape_collections = self.scrape_ratings and settings.getbool(
            "SCRAPE_BGG_COLLECTIONS"
        )
        self.scrape_users = self.scrape_ratings and settings.getbool("SCRAPE_BGG_USERS")
        self.min_votes = settings.getint("MIN_VOTES", self.min_votes)

        self.logger.info("scrape ratings: %r", self.scrape_ratings)
        self.logger.info("scrape collections: %r", self.scrape_collections)
        self.logger.info("scrape users: %r", self.scrape_users)

    def _spider_opened(self):
        state = getattr(self, "state", None)

        if state is None:
            self.logger.warning("no spider state found")
            state = {}
            self.state = state

        ids_seen = state.get("ids_seen") or frozenset()
        self.logger.info("%d ID(s) seen in previous state", len(ids_seen))

        self._ids_seen |= ids_seen

        self.state["ids_seen"] = self._ids_seen

    def _api_url(self, action, **kwargs):
        kwargs["pagesize"] = self.page_size
        params = ((k, v) for k, v in kwargs.items() if k and v is not None)
        return "{}/{}?{}".format(
            self.xml_api_url, action, urlencode(sorted(params, key=lambda x: x[0]))
        )

    def _game_requests(self, *bgg_ids, batch_size=10, page=1, priority=0, **kwargs):
        bgg_ids = clear_list(map(parse_int, bgg_ids))

        if not bgg_ids:
            return

        bgg_ids = (
            (bgg_id for bgg_id in bgg_ids if bgg_id not in self._ids_seen)
            if page == 1
            else bgg_ids
        )

        for batch in batchify(bgg_ids, batch_size):
            batch = tuple(batch)

            ids = ",".join(map(str, batch))

            url = (
                self._api_url(
                    action="thing",
                    id=ids,
                    stats=1,
                    videos=1,
                    versions=int(self.scrape_ratings),
                    ratingcomments=int(self.scrape_ratings),
                    page=1,
                )
                if page == 1
                else self._api_url(
                    action="thing", id=ids, versions=1, ratingcomments=1, page=page
                )
            )

            request = Request(url, callback=self.parse_game, priority=priority)

            if len(batch) == 1:
                request.meta["bgg_id"] = batch[0]
            request.meta["page"] = page
            request.meta.update(kwargs)

            yield request

            if page == 1:
                self._ids_seen.update(batch)

    def _game_request(self, bgg_id, default=None, **kwargs):
        return next(self._game_requests(bgg_id, **kwargs), default)

    def collection_request(self, user_name, *, meta=None, played=None, **kwargs):
        """ make a collection request for that user """

        user_name = user_name.lower()
        url = self._api_url(
            action="collection",
            username=user_name,
            subtype="boardgame",
            excludesubtype="boardgameexpansion",
            stats=1,
            version=0,
            played=played,
        )

        request = Request(url, callback=self.parse_collection, **kwargs)
        if meta:
            request.meta.update(meta)
        request.meta["bgg_user_name"] = user_name

        return request

    def _filter_votes(self, votes_best, votes_rec, votes_not, best=False):
        if votes_best + votes_rec + votes_not < self.min_votes / 2:
            return False

        votes_true = votes_best
        votes_false = votes_not
        if best:
            votes_false += votes_rec
        else:
            votes_true += votes_rec

        return votes_true > votes_false

    def _player_count_votes(self, game):
        min_players = _parse_int(game, "minplayers/@value")
        max_players = _parse_int(game, "maxplayers/@value")

        polls = game.xpath('poll[@name = "suggested_numplayers"]')
        poll = polls[0] if polls else None

        if not poll or _parse_int(poll, "@totalvotes", 0) < self.min_votes:
            return min_players, max_players, min_players, max_players

        votes = sorted(_parse_player_count(poll), key=lambda x: x[0])
        recommended = [
            vote[0] for vote in votes if self._filter_votes(*vote[1:], best=False)
        ]
        best = [vote[0] for vote in votes if self._filter_votes(*vote[1:], best=True)]

        return (
            min(recommended, default=min_players),
            max(recommended, default=max_players),
            min(best, default=min_players),
            max(best, default=max_players),
        )

    def _poll(
        self, game, name, attr="value", enum=False, func=statistics.mean, default=None
    ):
        polls = game.xpath('poll[@name = "{}"]'.format(name))
        poll = polls[0] if polls else None

        if not poll or _parse_int(poll, "@totalvotes", 0) < self.min_votes:
            return default

        try:
            return func(_parse_votes(poll, attr, enum))
        except Exception as exc:
            self.logger.exception(exc)

        return default

    def _user_item_or_request(self, user_name, priority=2, **kwargs):
        if not user_name:
            return None

        user_name = user_name.lower()
        kwargs.setdefault("scraped_at", now())

        ldr = UserLoader(item=UserItem(bgg_user_name=user_name))
        for field_name, value in kwargs.items():
            ldr.add_value(field_name, value)
        item = ldr.load_item()

        if not self.scrape_users:
            return item

        url = self._api_url(action="user", name=user_name)
        return Request(
            url=url,
            callback=partial(self.parse_user, item=item),
            meta={"item": item},
            priority=priority,
        )

    def parse(self, response):
        """
        @url https://boardgamegeek.com/browse/boardgame/
        @returns items 0 0
        @returns requests 11
        """

        next_page = response.xpath('//a[@title = "next page"]/@href').extract_first()
        if next_page:
            yield Request(
                response.urljoin(next_page),
                callback=self.parse,
                priority=1,
                meta={"max_retry_times": 10},
            )

        urls = response.xpath("//@href").extract()
        bgg_ids = filter(None, map(extract_bgg_id, map(response.urljoin, urls)))
        yield from self._game_requests(*bgg_ids)

        user_names = filter(None, map(extract_bgg_user_name, urls))
        scraped_at = now()

        for user_name in clear_list(user_names):
            yield self.collection_request(
                user_name
            ) if self.scrape_collections else self._user_item_or_request(
                user_name, scraped_at=scraped_at
            )

    def parse_game(self, response):
        # pylint: disable=line-too-long
        """
        @url https://www.boardgamegeek.com/xmlapi2/thing?id=13,822,36218&stats=1&versions=1&videos=1&ratingcomments=1&page=1&pagesize=100
        @returns items 3 3
        @returns requests 303 303
        @scrapes name alt_name year description \
            designer artist publisher \
            url image_url video_url \
            min_players max_players min_players_rec max_players_rec \
            min_players_best max_players_best \
            min_age min_age_rec min_time max_time \
            game_type category mechanic cooperative compilation family expansion \
            rank num_votes avg_rating stddev_rating \
            bayes_rating worst_rating best_rating \
            complexity easiest_complexity hardest_complexity \
            language_dependency lowest_language_dependency highest_language_dependency \
            bgg_id scraped_at
        """

        profile_url = response.meta.get("profile_url")
        scraped_at = now()

        for game in response.xpath("/items/item"):
            bgg_id = parse_int(
                game.xpath("@id").extract_first() or response.meta.get("bgg_id")
            )
            page = parse_int(
                game.xpath("comments/@page").extract_first()
                or response.meta.get("page")
            )
            total_items = parse_int(
                game.xpath("comments/@totalitems").extract_first()
                or response.meta.get("total_items")
            )
            comments = game.xpath("comments/comment") if self.scrape_ratings else ()

            if (
                page is not None
                and total_items is not None
                and comments
                and page * self.page_size < total_items
            ):
                # pylint: disable=invalid-unary-operand-type
                yield self._game_request(
                    bgg_id,
                    page=page + 1,
                    priority=-page,
                    skip_game_item=True,
                    profile_url=profile_url,
                )

            for comment in comments:
                user_name = comment.xpath("@username").extract_first()

                if not user_name:
                    self.logger.warning("no user name found, cannot process rating")
                    continue

                user_name = user_name.lower()

                if self.scrape_collections:
                    yield self.collection_request(user_name)
                    continue

                yield self._user_item_or_request(user_name, scraped_at=scraped_at)

                ldr = RatingLoader(
                    item=RatingItem(
                        item_id=f"{user_name}:{bgg_id}",
                        bgg_id=bgg_id,
                        bgg_user_name=user_name,
                        scraped_at=scraped_at,
                    ),
                    selector=comment,
                    response=response,
                )
                ldr.add_xpath("bgg_user_rating", "@rating")
                ldr.add_xpath("comment", "@value")
                yield ldr.load_item()

            if response.meta.get("skip_game_item"):
                continue

            ldr = GameLoader(
                item=GameItem(
                    bgg_id=bgg_id,
                    scraped_at=scraped_at,
                    worst_rating=1,
                    best_rating=10,
                    easiest_complexity=1,
                    hardest_complexity=5,
                    lowest_language_dependency=1,
                    highest_language_dependency=5,
                ),
                selector=game,
                response=response,
            )

            ldr.add_xpath("name", 'name[@type = "primary"]/@value')
            ldr.add_xpath("alt_name", "name/@value")
            ldr.add_xpath("year", "yearpublished/@value")
            ldr.add_xpath("description", "description")

            ldr.add_value(
                "designer", _value_id(game.xpath('link[@type = "boardgamedesigner"]'))
            )
            ldr.add_value(
                "artist", _value_id(game.xpath('link[@type = "boardgameartist"]'))
            )
            ldr.add_value(
                "publisher", _value_id(game.xpath('link[@type = "boardgamepublisher"]'))
            )

            ldr.add_value("url", profile_url)
            ldr.add_value(
                "url", "https://boardgamegeek.com/boardgame/{}".format(bgg_id)
            )
            images = game.xpath("image/text()").extract()
            ldr.add_value("image_url", (response.urljoin(i) for i in images))
            images = game.xpath("thumbnail/text()").extract()
            ldr.add_value("image_url", (response.urljoin(i) for i in images))
            videos = game.xpath("videos/video/@link").extract()
            ldr.add_value("video_url", (response.urljoin(v) for v in videos))

            (
                min_players_rec,
                max_players_rec,
                min_players_best,
                max_players_best,
            ) = self._player_count_votes(game)

            ldr.add_xpath("min_players", "minplayers/@value")
            ldr.add_xpath("max_players", "maxplayers/@value")
            ldr.add_value("min_players_rec", min_players_rec)
            ldr.add_value("max_players_rec", max_players_rec)
            ldr.add_value("min_players_best", min_players_best)
            ldr.add_value("max_players_best", max_players_best)

            ldr.add_xpath("min_age", "minage/@value")
            ldr.add_xpath("max_age", "maxage/@value")
            ldr.add_value(
                "min_age_rec",
                self._poll(game, "suggested_playerage", func=statistics.median_grouped),
            )
            ldr.add_xpath("min_time", "minplaytime/@value")
            ldr.add_xpath("min_time", "playingtime/@value")
            ldr.add_xpath("max_time", "maxplaytime/@value")
            ldr.add_xpath("max_time", "playingtime/@value")
            ldr.add_xpath("max_time", "minplaytime/@value")

            ldr.add_value(
                "game_type",
                _value_id_rank(
                    game.xpath('statistics/ratings/ranks/rank[@type = "family"]')
                ),
            )
            ldr.add_value(
                "category", _value_id(game.xpath('link[@type = "boardgamecategory"]'))
            )
            ldr.add_value(
                "mechanic", _value_id(game.xpath('link[@type = "boardgamemechanic"]'))
            )
            # look for <link type="boardgamemechanic" id="2023" value="Co-operative Play" />
            ldr.add_value(
                "cooperative",
                bool(game.xpath('link[@type = "boardgamemechanic" and @id = "2023"]')),
            )
            ldr.add_value(
                "compilation",
                bool(
                    game.xpath(
                        'link[@type = "boardgamecompilation" and @inbound = "true"]'
                    )
                ),
            )
            ldr.add_xpath(
                "compilation_of",
                'link[@type = "boardgamecompilation" and @inbound = "true"]/@id',
            )
            ldr.add_value(
                "family", _value_id(game.xpath('link[@type = "boardgamefamily"]'))
            )
            ldr.add_value(
                "expansion", _value_id(game.xpath('link[@type = "boardgameexpansion"]'))
            )
            ldr.add_xpath(
                "implementation",
                'link[@type = "boardgameimplementation" and @inbound = "true"]/@id',
            )
            ldr.add_xpath("integration", 'link[@type = "boardgameintegration"]/@id')

            ldr.add_xpath(
                "rank", 'statistics/ratings/ranks/rank[@name = "boardgame"]/@value'
            )
            ldr.add_xpath("num_votes", "statistics/ratings/usersrated/@value")
            ldr.add_xpath("avg_rating", "statistics/ratings/average/@value")
            ldr.add_xpath("stddev_rating", "statistics/ratings/stddev/@value")
            ldr.add_xpath("bayes_rating", "statistics/ratings/bayesaverage/@value")
            ldr.add_xpath("complexity", "statistics/ratings/averageweight/@value")
            ldr.add_value(
                "language_dependency",
                self._poll(
                    game,
                    "language_dependence",
                    attr="level",
                    enum=True,
                    func=statistics.median_grouped,
                ),
            )

            yield ldr.load_item()

    def parse_collection(self, response):
        # pylint: disable=line-too-long
        """
        @url https://www.boardgamegeek.com/xmlapi2/collection?username=Markus+Shepherd&subtype=boardgame&excludesubtype=boardgameexpansion&stats=1&version=0
        @returns items 1000
        @returns requests 100
        @scrapes item_id bgg_id bgg_user_name bgg_user_owned bgg_user_prev_owned \
            bgg_user_for_trade bgg_user_want_to_play bgg_user_want_to_buy \
            bgg_user_preordered bgg_user_play_count updated_at scraped_at
        """

        user_name = response.meta.get("bgg_user_name") or extract_bgg_user_name(
            response.url
        )
        scraped_at = now()

        if not user_name:
            self.logger.warning("no user name found, cannot process collection")
            return

        user_name = user_name.lower()

        if not extract_query_param(response.url, "played"):
            updated_at = response.xpath("/items/@pubdate").extract_first()
            yield self._user_item_or_request(
                user_name, updated_at=updated_at, scraped_at=scraped_at
            )

            # explicitly fetch played games (not part of collection by default)
            yield self.collection_request(user_name, played=1, priority=1)

        games = response.xpath("/items/item")
        bgg_ids = games.xpath("@objectid").extract()
        yield from self._game_requests(*bgg_ids)

        for game in games:
            bgg_id = parse_int(game.xpath("@objectid").extract_first())

            if not bgg_id:
                self.logger.warning("no BGG ID found, cannot process rating")
                continue

            ldr = RatingLoader(
                item=RatingItem(
                    bgg_id=bgg_id, bgg_user_name=user_name, scraped_at=scraped_at
                ),
                selector=game,
                response=response,
            )

            ldr.add_value("item_id", parse_int(game.xpath("@collid").extract_first()))
            ldr.add_value("item_id", f"{user_name}:{bgg_id}")

            ldr.add_xpath("bgg_user_rating", "stats/rating/@value")
            ldr.add_xpath("bgg_user_owned", "status/@own")
            ldr.add_xpath("bgg_user_prev_owned", "status/@prevowned")
            ldr.add_xpath("bgg_user_for_trade", "status/@fortrade")
            ldr.add_xpath("bgg_user_want_in_trade", "status/@want")
            ldr.add_xpath("bgg_user_want_to_play", "status/@wanttoplay")
            ldr.add_xpath("bgg_user_want_to_buy", "status/@wanttobuy")
            ldr.add_xpath("bgg_user_preordered", "status/@preordered")
            ldr.add_xpath(
                "bgg_user_wishlist", 'status[@wishlist = "1"]/@wishlistpriority'
            )
            ldr.add_xpath("bgg_user_play_count", "numplays/text()")

            ldr.add_xpath("comment", "comment/text()")

            ldr.add_xpath("updated_at", "status/@lastmodified")

            yield ldr.load_item()

    # pylint: disable=no-self-use
    def parse_user(self, response, item=None):
        """
        @url https://www.boardgamegeek.com/xmlapi2/user?name=Markus+Shepherd
        @returns items 1 1
        @returns requests 0 0
        @scrapes item_id bgg_user_name first_name last_name registered last_login \
            country external_link image_url scraped_at
        """

        item = extract_item(item, response, UserItem)

        ldr = UserLoader(item=item, selector=response.xpath("/user"))

        ldr.add_xpath("item_id", "@id")

        ldr.add_xpath("bgg_user_name", "@name")
        ldr.add_xpath("first_name", "firstname/@value")
        ldr.add_xpath("last_name", "lastname/@value")

        ldr.add_xpath("registered", "yearregistered/@value")
        ldr.add_xpath("last_login", "lastlogin/@value")

        ldr.add_xpath("country", "country/@value")
        ldr.add_xpath("region", "stateorprovince/@value")

        ldr.add_xpath("external_link", "webaddress/@value")
        ldr.add_xpath("image_url", "avatarlink/@value")

        ldr.replace_value("scraped_at", now())

        return ldr.load_item()



# -*- coding: utf-8 -*-

""" util functions """

import json
import logging
import os
import re

from datetime import datetime, timezone
from functools import partial
from types import GeneratorType
from typing import Any, Dict, Iterable, List, Optional, Pattern, Union
from urllib.parse import ParseResult, parse_qs, unquote_plus, urlparse

from pytility import (
    arg_to_iter,
    clear_list,
    normalize_space,
    parse_float,
    parse_int,
    take_first,
    to_str,
    parse_date,
)
from scrapy.item import BaseItem, Item
from w3lib.html import replace_entities

LOGGER = logging.getLogger(__name__)

REGEX_ENTITIES = re.compile(r"(&#(\d+);)+")
REGEX_SINGLE_ENT = re.compile(r"&#(\d+);")

REGEX_BGG_ID = re.compile(r"^/(board)?game/(\d+).*$")
REGEX_BGG_USER = re.compile(r"^/user/([^/]+).*$")
REGEX_WIKIDATA_ID = re.compile(r"^/(wiki|entity|resource)/Q(\d+).*$")
REGEX_DBPEDIA_DOMAIN = re.compile(r"^[a-z]{2}\.dbpedia\.org$")
REGEX_DBPEDIA_ID = re.compile(r"^/(resource|page)/(.+)$")
REGEX_LUDING_ID = re.compile(r"^.*gameid/(\d+).*$")
REGEX_SPIELEN_ID = re.compile(
    r"^/(alle-brettspiele|messeneuheiten|ausgezeichnet-\d+)/(\w[^/]*).*$"
)
REGEX_FREEBASE_ID = re.compile(r"^/ns/(g|m)\.([^/]+).*$")
REGEX_BGA_ID = re.compile(r"^.*/game/([a-zA-Z0-9]+)(/.*)?$")


def to_lower(string):
    """ safely convert to lower case string, else return None """
    string = to_str(string)
    return string.lower() if string is not None else None


def identity(obj: Any) -> Any:
    """ do nothing """
    return obj


def _replace_utf_entities(match):
    try:
        values = tuple(map(parse_int, REGEX_SINGLE_ENT.findall(match.group(0))))
        bytes_ = bytes(values) if all(values) else None
        return bytes_.decode() if bytes_ else match.group(0)
    except Exception:
        pass
    return match.group(0)


def replace_utf_entities(string):
    """ replace XML entities weirdly encoded """
    return REGEX_ENTITIES.sub(_replace_utf_entities, string)


def replace_all_entities(string):
    """ replace all XML entities, even poorly encoded """
    # hack because BGG encodes 'Ü' as '&amp;#195;&amp;#156;' (d'oh!)
    # note that this may corrupt text that's actually encoded correctly!
    return replace_entities(
        replace_utf_entities(
            string.replace("&amp;", "&").replace("&amp;", "&").replace("&amp;", "&")
        )
    )


def extract_query_param(url: Union[str, ParseResult], field: str) -> Optional[str]:
    """ extract a specific field from URL query parameters """

    url = urlparse(url) if isinstance(url, str) else url
    query = parse_qs(url.query)
    values = query.get(field)

    return values[0] if values else None


def now(tzinfo=None):
    """ current time in UTC or given timezone """

    result = datetime.utcnow().replace(microsecond=0, tzinfo=timezone.utc)
    return result if tzinfo is None else result.astimezone(tzinfo)


def serialize_date(date, tzinfo=None):
    """seralize a date into ISO format if possible"""

    parsed = parse_date(date, tzinfo)
    return parsed.strftime("%Y-%m-%dT%T%z") if parsed else str(date) if date else None


def parse_json(file_or_string, **kwargs):
    """safely parse JSON string"""

    if file_or_string is None:
        return None

    try:
        return json.load(file_or_string, **kwargs)
    except Exception:
        pass

    try:
        return json.loads(to_str(file_or_string), **kwargs)
    except Exception:
        pass

    return None


def _json_default(obj):
    if isinstance(obj, BaseItem):
        return dict(obj)
    if isinstance(obj, (set, frozenset, range, GeneratorType)) or hasattr(
        obj, "__iter__"
    ):
        return list(obj)
    if isinstance(obj, datetime):
        return serialize_date(obj)
    return repr(obj)


def serialize_json(obj, file=None, **kwargs):
    """
    safely serialze JSON, turning iterables into lists, dates into ISO strings,
    and everything else into their representation
    """

    kwargs.setdefault("default", _json_default)

    if isinstance(file, (str, bytes)):
        LOGGER.info("opening file <%s> and writing JSON content", file)

        path_dir = os.path.abspath(os.path.split(file)[0])
        os.makedirs(path_dir, exist_ok=True)

        try:
            from smart_open import smart_open
        except ImportError:
            smart_open = open

        with smart_open(file, "w") as json_file:
            return json.dump(obj, json_file, **kwargs)

    if file is not None:
        LOGGER.debug("writing JSON content to opened file pointer <%s>", file)
        return json.dump(obj, file, **kwargs)

    return json.dumps(obj, **kwargs)


def str_to_parser(string):
    """ parser from key string """
    string = to_lower(string)
    if not string:
        return to_str
    return (
        parse_int
        if string == "int"
        else parse_float
        if string == "float"
        else partial(parse_date, tzinfo=timezone.utc)
        if string == "date"
        else to_lower
        if string in ("istr", "istring")
        else to_str
    )


def validate_range(value, lower=None, upper=None):
    """ validate that the given value is between lower and upper """
    try:
        if (lower is None or lower <= value) and (upper is None or value <= upper):
            return value
    except TypeError:
        pass
    return None


def json_from_response(response):
    """Parse JSON from respose if possible."""
    result = parse_json(response.text) if hasattr(response, "text") else None
    return result or {}


def extract_meta(response=None):
    """Extract meta object from response if possible."""
    if hasattr(response, "meta") and response.meta:
        return response.meta
    if hasattr(response, "request") and hasattr(response.request, "meta"):
        return response.request.meta or {}
    return {}


def extract_item(item=None, response=None, item_cls=Item):
    """Extract item from response if possible."""
    if item:
        return item
    meta = extract_meta(response)
    return meta.get("item") or item_cls()


def extract_url(item=None, response=None, default=None):
    """Extract URL from response if possible."""
    if item and item.get("url"):
        return item["url"]
    meta = extract_meta(response)
    if meta.get("url"):
        return meta["url"]
    if hasattr(response, "url") and response.url:
        return response.url
    if hasattr(response, "request") and hasattr(response.request, "url"):
        return response.request.url
    return default


def _match(string: str, comparison: Union[str, Pattern]):
    return (
        string == comparison
        if isinstance(comparison, str)
        else bool(comparison.match(string))
    )


def parse_url(
    url: Union[str, ParseResult, None],
    hostnames: Optional[Iterable[Union[str, Pattern]]] = None,
) -> Optional[ParseResult]:
    """ parse URL and optionally filter for hosts """
    url = urlparse(url) if isinstance(url, str) else url
    hostnames = tuple(arg_to_iter(hostnames))
    return (
        url
        if url
        and url.hostname
        and url.path
        and (
            not hostnames
            or any(_match(url.hostname, hostname) for hostname in hostnames)
        )
        else None
    )


def validate_url(
    url: Union[str, ParseResult, None],
    hostnames: Optional[Iterable[Union[str, Pattern]]] = None,
    schemes: Optional[Iterable[Union[str, Pattern]]] = None,
) -> Optional[str]:
    """Returns cleaned up URL iff valid with scheme, hostname, and path."""
    url = parse_url(url=url, hostnames=hostnames)
    schemes = frozenset(arg_to_iter(schemes))
    return (
        url.geturl()
        if url is not None and url.scheme and (not schemes or url.scheme in schemes)
        else None
    )


def extract_bgg_id(url: Union[str, ParseResult, None]) -> Optional[int]:
    """ extract BGG ID from URL """
    url = parse_url(url, ("boardgamegeek.com", "www.boardgamegeek.com"))
    if not url:
        return None
    match = REGEX_BGG_ID.match(url.path)
    bgg_id = parse_int(match.group(2)) if match else None
    return bgg_id if bgg_id is not None else parse_int(extract_query_param(url, "id"))


def extract_bgg_user_name(url: Union[str, ParseResult, None]) -> Optional[str]:
    """ extract BGG user name from url """
    url = parse_url(url, ("boardgamegeek.com", "www.boardgamegeek.com"))
    if not url:
        return None
    match = REGEX_BGG_USER.match(url.path)
    user_name = (
        unquote_plus(match.group(1)) if match else extract_query_param(url, "username")
    )
    return user_name.lower() if user_name else None


def extract_wikidata_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """ extract Wikidata ID from URL """
    url = parse_url(url, ("wikidata.org", "www.wikidata.org", "wikidata.dbpedia.org"))
    if not url:
        return None
    match = REGEX_WIKIDATA_ID.match(url.path)
    return f"Q{match.group(2)}" if match else extract_query_param(url, "id")


def extract_wikipedia_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """ extract Wikipedia ID from URL """
    url = parse_url(url, ("en.wikipedia.org", "en.m.wikipedia.org"))
    return (
        unquote_plus(url.path[6:]) or None
        if url and url.path.startswith("/wiki/")
        else None
    )


def extract_dbpedia_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """ extract DBpedia ID from URL """
    url = parse_url(url, ("dbpedia.org", "www.dbpedia.org", REGEX_DBPEDIA_DOMAIN))
    if not url:
        return None
    match = REGEX_DBPEDIA_ID.match(url.path)
    return unquote_plus(match.group(2)) if match else extract_query_param(url, "id")


def extract_luding_id(url: Union[str, ParseResult, None]) -> Optional[int]:
    """ extract Luding ID from URL """
    url = parse_url(url, ("luding.org", "www.luding.org"))
    if not url:
        return None
    match = REGEX_LUDING_ID.match(url.path)
    return (
        parse_int(match.group(1))
        if match
        else parse_int(extract_query_param(url, "gameid"))
    )


def extract_spielen_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """ extract Spielen.de ID from URL """
    url = parse_url(
        url, ("gesellschaftsspiele.spielen.de", "www.gesellschaftsspiele.spielen.de")
    )
    if not url:
        return None
    match = REGEX_SPIELEN_ID.match(url.path)
    spielen_id = unquote_plus(match.group(2)) if match else None
    return (
        spielen_id if parse_int(spielen_id) is None else extract_query_param(url, "id")
    )


def extract_freebase_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """ extract Freebase ID from URL """
    url = parse_url(url, ("rdf.freebase.com", "freebase.com"))
    if not url:
        return None
    match = REGEX_FREEBASE_ID.match(url.path)
    return (
        f"/{match.group(1)}/{match.group(2)}"
        if match
        else extract_query_param(url, "id")
    )


def extract_bga_id(url: Union[str, ParseResult, None]) -> Optional[str]:
    """ extract Board Game Atlas ID from URL """
    url = parse_url(url, ("boardgameatlas.com", "www.boardgameatlas.com"))
    if not url:
        return None
    match = REGEX_BGA_ID.match(url.path)
    if match:
        return match.group(1)
    ids_str = extract_query_param(url, "ids")
    ids = ids_str.split(",") if ids_str else ()
    return take_first(map(normalize_space, ids)) or extract_query_param(url, "game-id")


def extract_ids(*urls: Optional[str]) -> Dict[str, List[Union[int, str]]]:
    """ extract all possible IDs from all the URLs """
    urls = tuple(map(urlparse, urls))
    return {
        "bgg_id": clear_list(map(extract_bgg_id, urls)),
        "freebase_id": clear_list(map(extract_freebase_id, urls)),
        "wikidata_id": clear_list(map(extract_wikidata_id, urls)),
        "wikipedia_id": clear_list(map(extract_wikipedia_id, urls)),
        "dbpedia_id": clear_list(map(extract_dbpedia_id, urls)),
        "luding_id": clear_list(map(extract_luding_id, urls)),
        "spielen_id": clear_list(map(extract_spielen_id, urls)),
        "bga_id": clear_list(map(extract_bga_id, urls)),
    }



# -*- coding: utf-8 -*-

""" Wikidata spider """

import json

from urllib.parse import urlencode

from pytility import batchify, normalize_space
from scrapy import Request, Spider
from scrapy.loader.processors import MapCompose
from scrapy.utils.misc import arg_to_iter

from ..items import GameItem
from ..loaders import GameJsonLoader
from ..utils import (
    extract_ids,
    extract_wikidata_id,
    identity,
)


class WikidataSpider(Spider):
    """ Wikidata spider """

    name = "wikidata"
    allowed_domains = ("wikidata.org",)
    item_classes = (GameItem,)

    sparql_api_url = "https://query.wikidata.org/sparql"
    entity_data_url = (
        "https://www.wikidata.org/wiki/Special:EntityData/{wikidata_id}.{fformat}"
    )

    custom_settings = {
        "DOWNLOAD_DELAY": 10,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 4,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 2,
    }

    game_types = (
        "Q131436",  # board game
        "Q11410",  # game
        "Q142714",  # card game
        "Q573573",  # abstract strategy game
        "Q839864",  # party game
        "Q734698",  # collectible card game
        "Q788553",  # German-style board game
        "Q1191150",  # trick-taking game
        "Q1272194",  # tile-based game
        "Q1150710",  # strategy game
        "Q532716",  # miniature wargaming
        "Q1368898",  # game of skill
        "Q1509934",  # children's game
        "Q1311927",  # railway game
        "Q1515156",  # dice game
        "Q3244175",  # tabletop game
        "Q4927217",  # block wargame
        "Q2347716",  # paper-and-pencil game
        "Q6983400",  # Nazi board games
        "Q5161688",  # connection game
        "Q1783817",  # cooperative board game
        "Q10927576",  # board wargame
        "Q15220419",  # word game
        "Q14947863",  # parlour game
        "Q5249796",  # dedicated deck card game
        "Q15804899",  # Deck-building game
        "Q21608615",  # economic simulation board game
        "Q28807042",  # game variant
    )

    custom_settings = {
        "ROBOTSTXT_OBEY": False,
        "RESOLVE_LABEL_URL": entity_data_url.format(wikidata_id="{}", fformat="json"),
        "RESOLVE_LABEL_FIELDS": ("designer", "artist", "publisher"),
        "RESOLVE_LABEL_LANGUAGE_PRIORITIES": ("en",),
    }

    def _api_url(self, query):
        args = {"format": "xml", "query": query}
        return "{}?{}".format(self.sparql_api_url, urlencode(args))

    def _entity_url(self, wikidata_id, fformat="json"):
        return self.entity_data_url.format(wikidata_id=wikidata_id, fformat=fformat)

    def _type_requests(self, types, batch_size=10):
        query_tmpl = normalize_space(
            """
            SELECT DISTINCT ?game WHERE {{
                ?game <http://www.wikidata.org/prop/direct/P31> ?type .
                VALUES ?type {{ {} }}
            }}"""
        )

        if not batch_size:
            query = query_tmpl.format(" ".join(types))
            # self.logger.debug(query)
            yield Request(
                self.sparql_api_url,
                method="POST",
                body=urlencode({"query": query}),
                callback=self.parse_games,
                priority=1,
            )
            return

        for batch in batchify(types, batch_size):
            query = query_tmpl.format(" ".join(batch))
            # self.logger.debug(query)
            yield Request(self._api_url(query), callback=self.parse_games, priority=1)

    def start_requests(self):
        """ generate start requests """

        types = getattr(self, "game_types", None)

        if types:
            yield from self._type_requests(
                "<http://www.wikidata.org/entity/{}>".format(t) for t in types
            )
            return

        query = normalize_space(
            """
            SELECT DISTINCT ?type WHERE {
                ?game <http://www.wikidata.org/prop/direct/P2339> ?bgg ;
                      <http://www.wikidata.org/prop/direct/P31> ?type .
            }"""
        )
        # self.logger.debug(query)
        yield Request(self._api_url(query), callback=self.parse, priority=2)

    def parse(self, response):
        # pylint: disable=line-too-long
        """
        @url https://query.wikidata.org/sparql?format=xml&query=SELECT+DISTINCT+%3Ftype+WHERE+%7B+%3Fgame+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2Fdirect%2FP2339%3E+%3Fbgg+%3B+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2Fdirect%2FP31%3E+%3Ftype+.+%7D
        @returns items 0 0
        @returns requests 5
        """

        response.selector.register_namespace(
            "sparql", "http://www.w3.org/2005/sparql-results#"
        )
        types = response.xpath(
            '//sparql:binding[@name = "type"]/sparql:uri/text()'
        ).extract()
        self.logger.info("received %d types", len(types))
        yield from self._type_requests("<{}>".format(t) for t in types)

    def parse_games(self, response):
        # pylint: disable=line-too-long
        """
        @url https://query.wikidata.org/sparql?format=xml&query=SELECT+DISTINCT+%3Fgame+WHERE+%7B+%3Fgame+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fprop%2Fdirect%2FP31%3E+%3Ftype+.+VALUES+%3Ftype+%7B+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ131436%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ11410%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ142714%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ573573%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ839864%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ734698%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ788553%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ1191150%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ1272194%3E+%3Chttp%3A%2F%2Fwww.wikidata.org%2Fentity%2FQ1150710%3E+%7D+%7D
        @returns items 0 0
        @returns requests 3000
        """

        response.selector.register_namespace(
            "sparql", "http://www.w3.org/2005/sparql-results#"
        )

        games = response.xpath(
            '//sparql:binding[@name = "game"]/sparql:uri/text()'
        ).extract()

        self.logger.info("received %d games", len(games))

        for game in games:
            wikidata_id = extract_wikidata_id(game)
            if wikidata_id:
                yield Request(self._entity_url(wikidata_id), callback=self.parse_game)

    def parse_game(self, response):
        """
        @url https://www.wikidata.org/wiki/Special:EntityData/Q17271.json
        @returns items 1 1
        @returns requests 0 0
        @scrapes name alt_name designer publisher url official_url image_url external_link \
            min_players max_players bgg_id wikidata_id wikipedia_id freebase_id luding_id bga_id
        """

        try:
            result = json.loads(response.text)
        except Exception as exc:
            self.logger.warning(exc)
            return

        for game in result.get("entities", {}).values():
            ldr = GameJsonLoader(item=GameItem(), json_obj=game, response=response)

            ldr.add_jmes("name", "labels.en.value")
            ldr.add_jmes("name", "aliases.en[].value")
            ldr.add_jmes("name", "labels.*.value")
            ldr.add_jmes("name", "aliases.*[].value")
            ldr.add_jmes("alt_name", "labels.*.value")
            ldr.add_jmes("alt_name", "aliases.*[].value")
            # TODO parse time to year
            ldr.add_jmes("year", "claims.P577[].mainsnak.datavalue.value.time")
            # TODO P571 inception

            ldr.add_jmes(
                "designer", "claims.P178[].mainsnak.datavalue.value.id"
            )  # developer
            ldr.add_jmes(
                "designer", "claims.P50[].mainsnak.datavalue.value.id"
            )  # author
            ldr.add_jmes(
                "designer", "claims.P170[].mainsnak.datavalue.value.id"
            )  # creator
            ldr.add_jmes(
                "designer", "claims.P287[].mainsnak.datavalue.value.id"
            )  # designed by
            ldr.add_jmes(
                "artist", "claims.P110[].mainsnak.datavalue.value.id"
            )  # illustrator
            ldr.add_jmes("publisher", "claims.P123[].mainsnak.datavalue.value.id")

            ldr.add_value("url", response.url)
            ldr.add_jmes(
                "image_url",
                "claims.P18[].mainsnak.datavalue.value",
                MapCompose(identity, response.urljoin),
            )
            # official website
            ldr.add_jmes("official_url", "claims.P856[].mainsnak.datavalue.value")
            # Wikipedia pages
            ldr.add_jmes("external_link", "sitelinks.*.url")

            ldr.add_jmes(
                "min_players", "claims.P1872[].mainsnak.datavalue.value.amount"
            )
            ldr.add_jmes(
                "max_players", "claims.P1873[].mainsnak.datavalue.value.amount"
            )
            ldr.add_jmes("min_age", "claims.P2899[].mainsnak.datavalue.value.amount")
            ldr.add_jmes("max_age", "claims.P4135[].mainsnak.datavalue.value.amount")
            # TODO duration = P2047

            ldr.add_jmes("bgg_id", "claims.P2339[].mainsnak.datavalue.value")
            ldr.add_jmes("freebase_id", "claims.P646[].mainsnak.datavalue.value")
            ldr.add_jmes("wikidata_id", "id")
            ldr.add_jmes("wikidata_id", "title")
            ldr.add_jmes("luding_id", "claims.P3528[].mainsnak.datavalue.value")
            ldr.add_jmes("bga_id", "claims.P6491[].mainsnak.datavalue.value")
            ldr.add_value(
                None,
                extract_ids(
                    response.url,
                    *arg_to_iter(ldr.get_output_value("external_link")),
                    *arg_to_iter(ldr.get_output_value("official_url")),
                ),
            )

            yield ldr.load_item()



# -*- coding: utf-8 -*-

""" split JSONL files """

import argparse
import json
import logging
import sys

from functools import partial

from pytility import batchify
from smart_open import smart_open

LOGGER = logging.getLogger(__name__)
FIELDS = frozenset(
    {
        "article_id",
        "url_canonical",
        "url_mobile",
        "url_amp",
        "url_thumbnail",
        "published_at",
        "title_short",
        "author",
        "description",
        "category",
        "keyword",
        "section_inferred",
        "country",
        "language",
        "source_name",
    }
)


def _filter_fields(item, fields=None, exclude_empty=False):
    return {
        k: v
        for k, v in item.items()
        if (not fields or k in fields) and (not exclude_empty or v)
    }


def _load_items(iterable, fields=None):
    if isinstance(iterable, str):
        with smart_open(iterable, "r") as file_obj:
            yield from _load_items(file_obj, fields=fields)
        return

    items = map(json.loads, iterable)
    items = map(partial(_filter_fields, fields=fields, exclude_empty=True), items)

    yield from items


def _parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("infile", help="input file")
    parser.add_argument("--batch", "-b", type=int, help="batch size")
    parser.add_argument("--outfile", "-o", help="output file path")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def _main():
    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    LOGGER.info("loading items from <%s>...", args.infile)

    items = tuple(_load_items(args.infile, fields=FIELDS))
    batches = batchify(items, args.batch) if args.batch else (items,)
    total = len(items)
    count = 0

    for i, batch in enumerate(batches):
        batch = list(batch)
        count += len(batch)
        result = {
            "count": total,
            "previous": i - 1 if i else None,
            "next": i + 1 if count < total else None,
            "results": batch,
        }

        if not args.outfile or args.outfile == "-":
            json.dump(result, sys.stdout, sort_keys=True)
            print()

        else:
            out_path = args.outfile.format(number=i)
            LOGGER.info("writing batch #%d to <%s>...", i, out_path)
            with smart_open(out_path, "w") as out_file:
                json.dump(result, out_file, sort_keys=True)

    LOGGER.info("done")


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

""" merge different sources """

import argparse
import json
import logging
import math
import os
import re
import sys

from collections import defaultdict
from functools import lru_cache
from itertools import chain
from pathlib import Path
from urllib.parse import urlparse
from pkg_resources import resource_stream

import dedupe
import yaml

from pytility import clear_list, parse_float, parse_int
from scrapy.utils.misc import arg_to_iter, load_object
from smart_open import smart_open

from .items import GameItem
from .utils import parse_json, serialize_json

LOGGER = logging.getLogger(__name__)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def abs_comp(field_1, field_2):
    """ returns absolute value of difference if both arguments are valid, else inf """
    field_1 = parse_float(field_1)
    field_2 = parse_float(field_2)
    return math.inf if field_1 is None or field_2 is None else abs(field_1 - field_2)


@lru_cache(8)
def _fields(resource="fields.yaml"):
    LOGGER.info("loading dedupe fields from <%s>", resource)
    fields = yaml.safe_load(resource_stream(__name__, resource))

    for field in fields:
        if field.get("comparator"):
            field["comparator"] = load_object(field["comparator"])

    return fields


DEDUPE_FIELDS = tuple(_fields())

VALUE_ID_REGEX = re.compile(r"^(.*?)(:(\d+))?$")
VALUE_ID_FIELDS = ("designer", "artist", "publisher")


def smart_exists(path, raise_exc=False):
    """ returns True if given path exists """

    url = urlparse(path)

    if url.scheme == "s3":
        try:
            import boto
        except ImportError as exc:
            LOGGER.error("<boto> library must be importable: %s", exc)
            if raise_exc:
                raise exc
            return False

        try:
            bucket = boto.connect_s3().get_bucket(url.hostname, validate=True)
            key = bucket.new_key(url.path[1:])
            return key.exists()

        except Exception as exc:
            LOGGER.error(exc)
            if raise_exc:
                raise exc

        return False

    try:
        return os.path.exists(url.path)

    except Exception as exc:
        LOGGER.error(exc)
        if raise_exc:
            raise exc

    return False


def _parse_value_id(string, regex=VALUE_ID_REGEX):
    match = regex.match(string) if string else None
    if not match or parse_int(match.group(3)) == 3:  # filter out '(Uncredited):3'
        return None
    return match.group(1) or None


def _parse_game(game):
    for field in DEDUPE_FIELDS:
        game.setdefault(field["field"], None)
        if field["type"] == "Set":
            game[field["field"]] = tuple(arg_to_iter(game[field["field"]])) or None
    game["names"] = tuple(
        clear_list(
            chain(arg_to_iter(game.get("name")), arg_to_iter(game.get("alt_name")))
        )
    )
    for field in VALUE_ID_FIELDS:
        game[field] = tuple(
            clear_list(map(_parse_value_id, arg_to_iter(game.get(field))))
        )
    return game


def _load_games(*args):
    for files in args:
        for file in arg_to_iter(files):
            if not file:
                continue

            LOGGER.info("reading from file <%s>", file)

            try:
                with smart_open(file, "r") as file_obj:
                    games = map(parse_json, file_obj)
                    games = filter(None, games)
                    games = map(GameItem.parse, games)
                    games = map(dict, games)
                    games = map(_parse_game, games)
                    yield from games

            except Exception:
                LOGGER.exception("there was an error reading from file <%s>", file)


def _make_id(game, id_field="id", id_prefix=None):
    id_value = game.get(id_field)
    return (
        None if not id_value else f"{id_prefix}:{id_value}" if id_prefix else id_value
    )


def _make_data(games, id_field="id", id_prefix=None):
    return {_make_id(game, id_field, id_prefix): game for game in games}


def _process_item(item):
    if isinstance(item, dict):
        return {k: _process_item(v) for k, v in item.items()}
    if isinstance(item, (frozenset, list, set, tuple)):
        item = type(item)(_process_item(v) for v in item)
    try:
        return dedupe.serializer._to_json(item)
    except TypeError:
        pass
    return item


def _process_game(game, fields=DEDUPE_FIELDS):
    return {field["field"]: game.get(field["field"]) for field in fields}


def _process_training(training):
    # training = {"distinct": [(item_1, item_2), ...], "match": [(item_1, item_2), ...]}
    training = {
        key: [[_process_game(game) for game in pair] for pair in pairs]
        for key, pairs in training.items()
    }
    return _process_item(training)


def _write_training(model, file_obj, **json_kwargs):
    training_pairs = _process_training(model.training_pairs)
    json.dump(obj=training_pairs, fp=file_obj, **json_kwargs)


def _train_gazetteer(
    data_1,
    data_2,
    fields=DEDUPE_FIELDS,
    training_file=None,
    manual_labelling=False,
    pretty_print=False,
):
    LOGGER.info("training gazetteer with fields: %r", fields)

    gazetteer = dedupe.Gazetteer(fields)

    if training_file and smart_exists(training_file):
        LOGGER.info("reading existing training from <%s>", training_file)
        with smart_open(training_file, "r") as file_obj:
            gazetteer.prepare_training(
                data_1=data_1, data_2=data_2, training_file=file_obj, sample_size=50_000
            )
    else:
        gazetteer.prepare_training(data_1=data_1, data_2=data_2, sample_size=50_000)

    if manual_labelling:
        LOGGER.info("start interactive labelling")
        dedupe.convenience.console_label(gazetteer)

    if training_file:
        LOGGER.info("write training data back to <%s>", training_file)
        with smart_open(training_file, "w") as file_obj:
            # bug in dedupe preventing training from being serialized correctly
            # gazetteer.write_training(file_obj)
            if pretty_print:
                _write_training(gazetteer, file_obj, sort_keys=True, indent=4)
            else:
                _write_training(gazetteer, file_obj)
        # if pretty_print:
        #     with smart_open(training_file, "r") as file_obj:
        #         training = parse_json(file_obj)
        #     with smart_open(training_file, "w") as file_obj:
        #         serialize_json(obj=training, file=file_obj, sort_keys=True, indent=4)

    LOGGER.info("done labelling, begin training")
    gazetteer.train(recall=0.9, index_predicates=True)

    gazetteer.cleanup_training()

    return gazetteer


def _extract_site(path):
    path = Path(path)
    return path.stem.split("_")[0] or None


def link_games(
    gazetteer,
    paths,
    id_prefixes=None,
    id_fields=None,
    training_file=None,
    manual_labelling=False,
    threshold=None,
    output=None,
    pretty_print=True,
):
    """ find links for games """

    paths = tuple(arg_to_iter(paths))
    if len(paths) < 2:
        raise ValueError(f"need at least 2 files to link games, but received {paths}")

    id_prefixes = tuple(arg_to_iter(id_prefixes))
    id_prefixes = id_prefixes + tuple(map(_extract_site, paths[len(id_prefixes) :]))
    id_fields = tuple(arg_to_iter(id_fields))
    id_fields = id_fields + tuple(
        f"{prefix}_id" for prefix in id_prefixes[len(id_fields) :]
    )

    games_canonical = _load_games(paths[0])
    data_canonical = _make_data(games_canonical, id_fields[0], id_prefixes[0])
    del games_canonical

    LOGGER.info(
        "loaded %d games in the canonical dataset <%s>", len(data_canonical), paths[0]
    )

    data_link = {}
    for path, id_field, id_prefix in zip(paths[1:], id_fields[1:], id_prefixes[1:]):
        games = _load_games(path)
        data_link.update(_make_data(games, id_field, id_prefix))
        del games

    LOGGER.info("loaded %d games to link in the datasets %s", len(data_link), paths[1:])

    if training_file:
        gazetteer_trained = _train_gazetteer(
            data_canonical,
            data_link,
            training_file=training_file,
            manual_labelling=manual_labelling,
            pretty_print=pretty_print,
        )

        if isinstance(gazetteer, str):
            LOGGER.info("saving gazetteer model to <%s>", gazetteer)
            with smart_open(gazetteer, "wb") as file_obj:
                gazetteer_trained.write_settings(file_obj)

        gazetteer = gazetteer_trained
        del gazetteer_trained

    elif isinstance(gazetteer, str):
        LOGGER.info("reading gazetteer model from <%s>", gazetteer)
        with smart_open(gazetteer, "rb") as file_obj:
            gazetteer = dedupe.StaticGazetteer(file_obj)

    gazetteer.index(data_canonical)

    LOGGER.info("using gazetteer model %r", gazetteer)

    threshold = threshold or 0.5  # TODO estimate threshold based on recall

    LOGGER.info("using threshold %.3f", threshold)

    clusters = gazetteer.search(
        data=data_link, threshold=threshold, n_matches=None, generator=True
    )
    del gazetteer

    links = defaultdict(set)
    for link_id, canonical_ids in clusters:
        for canonical_id, _ in canonical_ids:
            links[canonical_id].add(link_id)
    del clusters

    LOGGER.info("found links for %d items", len(links))

    if output == "-":
        for id_canonical, linked in links.items():
            LOGGER.info("%s <-> %s", id_canonical, linked)

    elif output:
        LOGGER.info("saving clusters as JSON to <%s>", output)
        links_sorted = {
            key: sorted(value) for key, value in links.items() if key and value
        }
        json_formats = {"sort_keys": True, "indent": 4} if pretty_print else {}
        with smart_open(output, "w") as file_obj:
            serialize_json(obj=links_sorted, file=file_obj, **json_formats)
        del links_sorted

    return links


def _parse_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "file_canonical", help="input JSON Lines files with canonical dataset"
    )
    parser.add_argument("files_link", nargs="+", help="input JSON Lines files to link")
    parser.add_argument("--id-fields", "-i", nargs="+", help="ID fields")
    parser.add_argument("--id-prefixes", "-I", nargs="+", help="ID prefixes")
    parser.add_argument("--train", "-t", action="store_true", help="train deduper")
    parser.add_argument(
        "--training-file",
        "-T",
        default=os.path.join(BASE_DIR, "cluster", "training.json"),
        help="training JSON file",
    )
    parser.add_argument(
        "--gazetteer-file",
        "-G",
        default=os.path.join(BASE_DIR, "cluster", "gazetteer.pickle"),
        help="gazetteer model file",
    )
    parser.add_argument("--threshold", "-r", type=float, help="clustering threshold")
    parser.add_argument("--output", "-o", help="output location")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def _main():
    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    link_games(
        gazetteer=args.gazetteer_file,
        paths=[args.file_canonical] + args.files_link,
        id_prefixes=args.id_prefixes,
        id_fields=args.id_fields,
        training_file=args.training_file if args.train else None,
        manual_labelling=args.train,
        threshold=args.threshold,
        output=args.output or "-",
    )


if __name__ == "__main__":
    _main()
